import Progress from '../util/Progress';
import { ClassMap, ClassMapEntity, Sources } from './Sources';
import fs from 'fs';
import { JSDOM, VirtualConsole } from 'jsdom';

class Analyzer {
	public static analyze(classMap: ClassMap, callback?: () => any): void {
		const script = fs.readFileSync(Sources.dzAppPath, 'utf8').toString();
		
		const virtualConsole = new VirtualConsole();
		virtualConsole.on('jsdomError', (error) => {});
		
		const dom = new JSDOM(`<!DOCTYPE html><html><head><meta http-equiv="Content-Type" content="text/html; charset=UTF-8"></head><body><script type="text/javascript">${script}</script></body></html>`, {
			virtualConsole: virtualConsole,
			runScripts: 'dangerously',
			resources: 'usable'
		});

		dom.window.document.addEventListener('DOMContentLoaded', () => {
			try {
				const Z8Classes = (dom.window as any).Z8.classes;

				const progress = Progress.start('Sources update, Step 4 [Analyze]:', Object.keys(Z8Classes).length + Object.keys(classMap).length + 1);

				Analyzer.filterZ8Classes(Z8Classes, progress);
				Analyzer.analyzeZ8Classes(Z8Classes, classMap, progress);

				fs.writeFileSync(Sources.classMapPath, JSON.stringify(classMap), { flag: 'a+', encoding: 'utf8' });

				progress.finish();

				if(callback)
					callback();
			} catch (e) {
				console.log(e);
			}

			dom.window.close();
		});
	}

	private static filterZ8Classes(Z8Classes: any, progress: Progress): any {
		const classesToDelete = [];
		for(const key of Object.keys(Z8Classes)) {
			const shortClassName = Analyzer.getShortClassName(Z8Classes[key]);
			if(key !== shortClassName && shortClassName != null)
				classesToDelete.push(shortClassName);
			progress.next();
		}

		progress.sumTotal(classesToDelete.length);

		for(const cls of classesToDelete) {
			delete Z8Classes[cls];
			progress.next();
		}
	}

	private static analyzeZ8Classes(Z8Classes: any, classMap: ClassMap, progress: Progress): void {
		for(const className of Object.keys(classMap)) {
			let cls = classMap[className];
			const z8Cls = Z8Classes[className];

			cls = Analyzer.setParentsBranch(cls, Analyzer.getParentsBranch(z8Cls, Z8Classes, classMap, [], null)!);
			cls = Analyzer.setExtends(cls, Analyzer.getSuperclassName(z8Cls));
			cls = Analyzer.setShortName(cls, Analyzer.getShortClassName(z8Cls));
			cls = Analyzer.setMixins(cls, Analyzer.getMixins(z8Cls, classMap));

			classMap[className] = cls;

			progress.next();
		}
	}

	private static getSuperclassName(z8Cls: any): string | null {
		return z8Cls.superclass.$className || null;
	}

	private static getClassName(z8Cls: any): string {
		return z8Cls.prototype.$className;
	}

	private static getShortClassName(z8Cls: any): string | null {
		const prototype = z8Cls.prototype;
		const superclass = z8Cls.superclass;

		return prototype.$shortClassName !== superclass.$shortClassName ? prototype.$shortClassName : null;
	}

	// TODO: remove previousSuperclassName when Z8.define will be fixed
	private static getParentsBranch(z8Cls: any, Z8Classes: any, classMap: ClassMap, parentsBranch: string[], previousSuperclassName: string | null): string[] | void {
		const clsName = Analyzer.getClassName(z8Cls);
		const superClsName = Analyzer.getSuperclassName(z8Cls);
		
		if(!superClsName || previousSuperclassName === superClsName)
			return parentsBranch.reverse();

		parentsBranch.push(superClsName);

		if(clsName !== superClsName)
			classMap[superClsName] = Analyzer.appendChild(classMap[superClsName], clsName);
		classMap[superClsName] = Analyzer.appendChildren(classMap[superClsName], parentsBranch.filter((parentClsName) => { return parentClsName !== superClsName; }));
		return Analyzer.getParentsBranch(Z8Classes[superClsName], Z8Classes, classMap, parentsBranch, superClsName);
	}

	private static getMixins(z8Cls: any, classMap: ClassMap): string[] {
		if(!z8Cls.prototype.$mixins)
			return [];
		return z8Cls.prototype.$mixins.map((mixin: any) => {
			const mixinClassName = Analyzer.getClassName(mixin);
			Analyzer.appendMixedIn(classMap[Analyzer.getClassName(mixin)], Analyzer.getClassName(z8Cls));
			return mixinClassName;
		});
	}

	// ClassMapEntity methods
	private static appendChild(cls: ClassMapEntity, childName: string): ClassMapEntity {
		if(!cls.children.includes(childName))
			cls.children.push(childName);
		return cls;
	}

	private static appendChildren(cls: ClassMapEntity, childNames: string[]): ClassMapEntity {
		for(const childName of childNames)
			Analyzer.appendChild(cls, childName);
		return cls;
	}

	private static setParentsBranch(cls: ClassMapEntity, parentsBranch: string[]): ClassMapEntity {
		cls.parentsBranch = parentsBranch;
		return cls;
	}

	private static setExtends(cls: ClassMapEntity, className: string | null): ClassMapEntity {
		cls.extends = className;
		return cls;
	}

	private static setShortName(cls: ClassMapEntity, shortName: string | null): ClassMapEntity {
		cls.shortName = shortName;
		return cls;
	}

	private static setMixins(cls: ClassMapEntity, mixins: string[]): ClassMapEntity {
		cls.mixins = mixins;
		return cls;
	}

	private static appendMixedIn(cls: ClassMapEntity, mixedTo: string): ClassMapEntity {
		if(!cls.mixedIn.includes(mixedTo))
			cls.mixedIn.push(mixedTo);
		return cls;
	}
}

export default Analyzer;